/**
 * Boid class for Processing, 
 * Nicolas Liautaud 2009
 * <br/>
 * Used by BoidsFlock class
 */

class Boid extends Element
{
  Boid(World world)
  {
    super(world);
    genes.add("cohesion", 0, 10);
    genes.add("separation", 0, 10);
    genes.add("imitation", 0, 10);
    genes.add("perception", 0, world.w);
  }
  Boid(Boid parent)
  {
    super(parent);
  }

  void update()
  { 
    float energyStart = genes.get("energyStart");
    float feedEffect = genes.get("feedEffect");
    
    // reproduction
    if(energy > 1.5 * energyStart && genes.get("reproductionAbility") > random(2))
    {
      new Boid(this);
      energy += genes.get("reproductionEffect");
    }
    
    // feed
    for(int i=0; i < world.size(); i++)
    {
      Element e = (Element)world.get(i);
      if(e instanceof Food)
      {
        // medium attraction when perceived
        effector(int(e.pos.x), int(e.pos.y), e.rad + genes.get("perception"), genes.get("feedAbility"));
        // big repulsion when touched
        effector(int(e.pos.x), int(e.pos.y), e.rad + rad, -20);
        // eat it at contact
        if(e.distance(this) <= rad + e.rad)
        {
          energy += feedEffect;
          e.energy -= feedEffect;
        }
      }
    }

    // do all other things from super class
    super.update();

    // boids rules
    keepDistance(childs, genes.get("separation") * 10, int(genes.get("separation")));
    cohesion(childs, genes.get("cohesion") - 5);
    matchVelocity(childs, genes.get("imitation") / 10);
  }

  /**
   * <p>Boid imitation : try to move in the same way than other boids.</p>
   * <p>Compute the average velocity of the percepted
   * flock, and add a part of that to velocity.</p>
   */
  void matchVelocity(ArrayList boids, float intensity)
  {
    PVector v = new PVector();
    int nbr = 0;
    for(int i=0; i < boids.size(); i++)
    { 
      Element b = (Element) boids.get(i);
      if(b != this && PVector.dist(pos, b.pos) < genes.get("perception"))
      {
        // add velocity of each other percepted boid
        v.add(b.vel);
        nbr++;
      }
    }
    if(nbr > 0)
    {
      v.div(nbr); // do average
      v.sub(vel);
      v.mult(intensity); // set intensity
      vel.add(v); // and add to velocity
    }
  }

  void limitVelocity(float lim){
    if(vel.mag() > lim)
      vel = PVector.mult(PVector.div(vel, vel.mag()), lim);
  }

  /**
   * Affect boid velocity with an linear point force.
   *
   * @param  x x-coordinate of the effector
   * @param  y y-coordinate of the effector
   * @param  a aura of the effector (effect distance)
   * @param  i intensity of the effector. > 0 for attraction, < 0 for repulsion
   */
  void effector(int x, int y, float a, float e)
  {
    float d = dist(x, y, pos.x, pos.y);
    if(d < a)
    {
      int f = 100;
      PVector v = new PVector();
      if(pos.x >= x) v.x += (a-d) * e / f;
      else v.x -= (a-d) * e / f;
      if(pos.y >= y) v.y += (a-d) * e / f;
      else v.y -= (a-d) * e / f;
      vel.sub(v);
    }
  }

  // add the vector display to the super class display.
  void display()
  {
    strokeWeight(2);
    super.display();
    PVector v = vel.get();
    v.normalize();
    v.mult(rad/2);
    strokeWeight(rad);
    stroke(col);
    point(pos.x + v.x, pos.y + v.y);
  }
}


