class Element implements Comparable
{
  World world;
  PVector pos, vel;
  float mass, energy, rad, age;
  ArrayList childs;
  color col;
  Genome genes;

  Element(World world)
  {
    this(world, world.randomPos()); 
  }
  Element(World w, PVector p)
  {
    world = w;
    pos = p;
    vel = new PVector(random(1), random(1));
    rad = random(4, 15);
    mass = random(rad/2, rad*2);
    col = color(random(0,255),0,random(0,255));

    genes = new Genome();
    float eS = genes.add("energyStart", 100, 1000);
    genes.add("health", -1, 0);
    genes.add("feedEffect", 0, 1);
    genes.add("feedAbility", 0, 1);
    genes.add("reproductionEffect", -eS, -eS/4);
    genes.add("reproductionAbility", 0, 1);

    energy = eS;
    age = 0;
    childs = new ArrayList();
  }
  /**
   * Element constructor derivated from parent element.
   */
  Element(Element parent)
  {
    world = parent.world;
    world.add(this);
    pos = PVector.add(parent.pos, new PVector(random(-1,1), random(-1,1)));
    vel = PVector.add(parent.vel, new PVector(random(-1,1), random(-1,1)));
    rad = parent.rad;
    mass = rad;
    col = parent.col;
    genes = parent.genes;

    energy = genes.get("energyStart");
    age = 0;
    childs = new ArrayList();
    parent.childs.add(this);
  }

  void update()
  {
    // life
    age += 1 / frameRate;
    energy += (age * genes.get("health")) / frameRate;

    // death
    if(energy < 1)
    {
      world.die(this);
    }

    // physics
    keepInWorld(world, 20);
    keepDistance(world.elements, 10, 1);
    keepDistance(childs, 20, 3);
    vel.mult(world.friction);
    vel.div(mass);
    pos.add(vel);
  }
  
  /**
   * <p>Element cohesion : try to approach other elements.</p>
   * <p>Compute the average direction to the barycenter of the
   * percepted flock, and add a part of that to velocity.</p>
   */
  void cohesion(ArrayList a, float intensity)
  {
    PVector v = new PVector();
    int nbr = 0;
    for(int i=0; i < a.size(); i++)
    {
      Element e = (Element)a.get(i);
      if(e != this && (e instanceof Food || (e instanceof Boid && this.distance(e) < genes.get("perception"))))
      {
        // add position of each other percepted boid
        v.add(e.pos);
        nbr++;
      }
    }
    if(nbr > 0)
    {
      v.div(nbr); // do average
      v.sub(pos);
      v.mult(intensity); // set intensity
      v.div(mass);
      vel.add(v); // and add to velocity
    }
  }

  /**
   * <p>Element avoidance : try to keep a minimum distance between others.</p>
   * <p>Compute the average opposite way to all the touched
   * elements, and add a part of that to velocity.</p>
   */
  void keepDistance(ArrayList a, float intensity, int m)
  {
    PVector u, v = new PVector();
    for(int i=0; i < a.size(); i++)
    {
      Element e = (Element)a.get(i);
      float d = distance(e);
      if(d != 0 && d < rad * m + e.rad)
      {
        u = PVector.sub(pos, e.pos);
        u.normalize();
        u.div(d);
        u.mult(e.mass);
        v.add(u);
      }
    }
    v.mult(intensity); // set intensity
    vel.add(v); // and add to velocity
  }

  /**
   * Keep the element in world by bouncing on sides.
   * i ~ 1.0 : hard
   * i ~ 0.1 : smooth
   */
  void keepInWorld(World world, float i)
  {
    int maxX = world.x + world.w;
    int maxY = world.y + world.h;
    PVector v = new PVector();
    // When a boid approach a side, an opposite vector is added to velocity
    if(pos.x <= world.x + rad)   v.x = (world.x + rad - pos.x) * i;
    else if(pos.x >= maxX - rad) v.x = (maxX - rad - pos.x) * i;
    if(pos.y <= world.y + rad)   v.y = (world.y + rad - pos.y) * i;
    else if(pos.y >= maxY - rad) v.y = (maxY - rad - pos.y) * i;
    vel.add(v);

  }

  void display()
  {
    fill(col, 50);
    stroke(col);
    ellipse(pos.x, pos.y, rad * 2, rad * 2);
    if(isUnderMouse() || world.selected == this) displayHover();
    if(displayLinks) displayLinks(world.elements);
  }
  void displayHover()
  {
    noFill();
    stroke(0);
    strokeWeight(1);
    //ellipse(pos.x, pos.y, rad * 2 + 20, rad * 2 + 20);
  }
  void displayLinks(ArrayList a)
  {
    if(a.indexOf(this) != -1)
    {
      stroke(col, 100);
      for(int i=0; i < childs.size(); i++)
      {
        Element child = (Element)childs.get(i);
        if(a.indexOf(child) != -1)
          line(pos.x, pos.y, child.pos.x, child.pos.y);
      }
    }
  }

  boolean isUnderMouse()
  {
    if(dist(pos.x, pos.y, mouseX, mouseY) < rad && !mousePressed) return true;
    else return false;
  }

  float distance(Element e)
  {
    return PVector.dist(pos, e.pos);
  }

  // sorting method
  int compareTo(Object other)
  {
    int n = 0;
    if(this.childs.size() > ((Element) other).childs.size()) n--;
    else n++;
    if(this.age > ((Element) other).age)
      n -= int((this.age - ((Element) other).age) / 50);
    else n += int((((Element) other).age - this.age) / 50);

    if(n == 0) return 0;
    else return n/abs(n);
  }
}















